import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.Font;
import java.awt.Toolkit;
import java.net.URL;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

/**
 * Ein Fenster zum Darstellen eines Bildes und eines zugehoerigen
 * Infotextes.
 * 
 * @author Bernd Gramlich 
 * @version 2013-06-20
 */
public class Bildfenster implements ActionListener
{
    private JFrame fenster;
    private JPanel fensterraum;
    private JLabel bildflaeche;
    private JTextArea textausgabe;
    private JScrollPane textrollflaeche;
    private JPanel knopfraum;
    private JButton schliessenknopf;
    private JFrame elternfenster;
    private WindowAdapter fensterBeobachter;

    /**
     * Konstruktor.
     * Erzeuge das Bildfenster mit dem Bild, dem Infotext und einem
     * Schliessenknopf darunter.
     * 
     * @param elter       Das Elternfenster dieses Bildfensters.
     * @param fenstername Der Name, der in der Kopfleiste dieses Bildfensters
     *                    stehen soll.
     * @param bildname    Der Name der Bilddatei, die dargestellt werden soll.
     *                    Wenn sie ungltig ist, wird kein Bild dargestellt.
     * @param infotext    Der Infotext, der unter dem Bild eingeblendet werden
     *                    soll. Dabei wird versucht, die Schriftgroesse so
     *                    anzupassen, dass der Infotext die Textausgabe
     *                    moeglichst genau fuellt.
     */
    public Bildfenster (JFrame elter, String fenstername, String bildname, String infotext)
    {
        /*
         * Deaktiviere das Elternfenster und die Leinwand.
         */
        elternfenster = elter;
        elternfenster.setEnabled (false);
        Leinwand.gibLeinwand ().gibFenster ().setEnabled (false);

        /*
         * Lade das Bild (sofern es existiert) und stelle es in der
         * Bildflaeche dar.
         */
        URL bildquelle = getClass ().getClassLoader ().getResource (bildname);

        if (null != bildquelle)
        {
            bildflaeche = new JLabel (new ImageIcon (bildquelle));
        }

        /*
         * Schreibe den Infotext in riesiger, fetter Festbreitenschrift auf
         * die Textausgabe. (Die Schriftgroesse wird spaeter noch angepasst.)
         * Verbiete Benutzereingaben in der Textausgabe. Sorge dafuer, dass
         * ein Zeilenumbruch stattfindet, und zwar bevorzugt an Wortgrenzen.
         */
        textausgabe = new JTextArea (infotext);
        textausgabe.setEnabled (false);
        textausgabe.setLineWrap (true);
        textausgabe.setWrapStyleWord (true);
        textausgabe.setFont (new Font (Font.MONOSPACED, Font.BOLD, 200));

        /*
         * Lege die Textausgabe auf eine Rollflaeche. Falls wider Erwarten
         * die Schriftgroessenberechnung schiefgeht, kann man dann mit dem
         * Schieberegler durch den Text rollen.
         */
        textrollflaeche = new JScrollPane (textausgabe);

        /*
         * Verwende den Schliessenknopf zum Schliessen des Bildfensters.
         */
        schliessenknopf = new JButton ("Schliessen");
        schliessenknopf.addActionListener (this);

        /*
         * Lege in den Fensterraum das Bild (sofern existent), darunter
         * den Text und ganz unten den Schliessenknopf.
         */
        fensterraum = new JPanel ();
        fensterraum.setLayout (new BorderLayout ());
        if (null != bildflaeche)
        {
            fensterraum.add (bildflaeche, BorderLayout.NORTH);
        }
        fensterraum.add (textrollflaeche, BorderLayout.CENTER);
        knopfraum = new JPanel ();
        knopfraum.add (schliessenknopf);
        fensterraum.add (knopfraum, BorderLayout.SOUTH);

        /*
         * Bestimme die Groesse des Bildschirms und die Groesse der
         * Bildflaeche.
         */
        Dimension bildschirmgroesse = Toolkit.getDefaultToolkit().getScreenSize();

        int bildbreite = 0;
        int bildhoehe  = 0;
        if (null != bildflaeche)
        {
            bildbreite = bildflaeche.getIcon ().getIconWidth ();
            bildhoehe  = bildflaeche.getIcon ().getIconHeight ();
        }

        /*
         * Erzeuge das Bildfenster. Sorge dafr, dass es etwas breiter und
         * deutlich hoeher als das Bild ist, aber immer noch auf den
         * Bildschirm passt.
         */
        fenster = new JFrame (fenstername);
        fenster.setMinimumSize (new Dimension ((2* bildbreite + (int) bildschirmgroesse.getWidth ()) / 3,
                (bildhoehe  + 2* (int) bildschirmgroesse.getHeight ()) / 3));
        fenster.setDefaultCloseOperation (JFrame.DO_NOTHING_ON_CLOSE);
        fenster.addWindowListener (new fensterHorcher ());
        fenster.getContentPane ().add (fensterraum);
        fenster.pack ();
        fenster.setLocationRelativeTo (null);
        fenster.setVisible (true);

        /*
         * Ermittle nun nachtraeglich die passende Schriftgroesse fuer den
         * Infotext, damit er die Textflaeche moeglichst genau fuellt.
         */
        setzeSchriftgroesse (infotext);
    }

    /**
     * Reagiere auf Benutzeraktionen.
     */
    public void actionPerformed (ActionEvent ereignis)
    {
        /*
         * Wenn der Schliessenknopf gedrueckt wird, loesche das Fenster und
         * aktiviere das Elternfenster und die Leinwand wieder.
         */
        if (ereignis.getSource () == schliessenknopf)
        {
            fenster.dispose ();
            elternfenster.setEnabled (true);
            elternfenster.toFront ();
            elternfenster.repaint ();
            Leinwand.gibLeinwand ().gibFenster ().setEnabled (true);
            Leinwand.gibLeinwand ().gibFenster ().requestFocus ();
        }
    }

    /**
     * Passe die Schriftgroesse der Textausgabe auf den enthaltenen
     * Text an. Momentan ist das noch eine rudimentaere Loesung mit
     * einer Festbreitenschrift.
     */
    private void setzeSchriftgroesse (String referenztext)
    {
        /*
         * Bestimme die Lnge des Referenztextes.
         */
        int textlaenge;
        
        if (null == referenztext)
        {
            textlaenge = 0;
        }
        else
        {
            textlaenge = referenztext.length ();
        }
        
        /*
         * Bestimme die Groesse der Textausgabe.
         */
        int breite     = textausgabe.getWidth  ();
        int hoehe      = textausgabe.getHeight ();
        
        /*
         * Berechne die Schriftgroesse. Vereinfachende Annahme dabei ist, dass bei
         * einer Festbreitenschrift der Gre g jeder Buchstabe in ein Kaestchen
         * der Groesse g mal g Bildpunkte passt. Das laesst ausser acht, dass Platz
         * durch den Zeilenumbruch verloren geht. Um das auszugleichen (und um
         * Division durch 0 zu vermeiden), steht im Nenner nicht textlaenge, sondern
         * textlaenge + 6. Wenn im Text viele lange Woerter einen unguenstigen
         * Zeilenumbruch erzwingen, kann es trotzdem passieren, dass die gewaehlte
         * Schriftgroesse zu gross ist und ein Rollbalken erscheint.
         */
        int groesse    = (int) Math.floor (Math.sqrt ((breite * hoehe) / (textlaenge + 6)));
        
        /*
         * Verwende fuer die Textausgabe eine fette Festbreitenschrift in der
         * berechneten Schriftgroesse.
         */
        textausgabe.setFont (new Font (Font.MONOSPACED, Font.BOLD, groesse));
    }

    /**
     * Die Klasse fensterHorcher dient dazu, dass das Fenster sich beim
     * Schliessen vom Betriebssystem her genauso verhaelt wie beim Druecken
     * des Schliessenknopfs.
     */
    private class fensterHorcher extends WindowAdapter
    {
        public void windowClosing (WindowEvent fensterereignis)
        {
            if (fensterereignis.getWindow () == fenster)
            {
                schliessenknopf.doClick ();
            }
        }
    }
}