import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.util.ArrayList;

/**
 * Liest ein Labyrinth aus einer Datei.
 * 
 * @author Bernd Gramlich
 * @author Romy B.
 * @author Stefanie E.
 * @author Nicole W.
 * @author Felicia W.
 * @version 2013-06-13
 */
public class Datenleser
{
    private static Datenleser datenleserSingleton;

    /*
     * Die Codezeichen, die wir im Labyrinth erwarten.
     */
    private final char WAND         = '#';
    private final char LEERRAUM     = '.';
    private final char EINGANG      = '@';
    private final char MARKENORD    = '^';
    private final char MARKEWEST    = '<';
    private final char MARKESUED    = 'v';
    private final char MARKEOST     = '>';
    private final char POSITIONNORD = 'N';
    private final char POSITIONSUED = 'S';
    private final char POSITIONWEST = 'W';
    private final char POSITIONOST  = 'O';
    private final char FEIND        = 'F';
    private final char ZEUGE1       = '1';
    private final char ZEUGE2       = '2';
    private final char ZEUGE3       = '3';
    private final char ZEUGE4       = '4';
    private final char ZEUGE5       = '5';
    private final char ZEUGE6       = '6';
    private final char ZEUGE7       = '7';
    private final char ZEUGE8       = '8';
    private final char ZEUGE9       = '9';
    private final char LEICHE       = 'L';
    private final char TAETER       = 'T';

    /*
     * Die Trennzeile im Labyrinth.
     */
    private final String TRENNZEILE = "-";

    /**
     * Konstruktor.
     * 
     * Der Konstruktor ist private, weil er von aussen
     * nicht aufgerufen werden soll. Stattdessen soll
     * mit gibDatenleser () auf den Datenleser
     * zugegriffen werden.
     */
    private Datenleser ()
    {
    }

    /**
     * Gib den einzigen verfuegbaren Datenleser zurueck
     * (also den "datenleserSingleton"). Wenn der
     * Datenleser noch nicht erzeugt ist, erzeuge ihn.
     */
    public static Datenleser gibDatenleser ()
    {
        if (null == datenleserSingleton)
        {
            datenleserSingleton = new Datenleser ();
        }

        return datenleserSingleton;
    }

    /**
     * Lies die vom Benutzer gewuenschte Labyrinthdatei ein
     * und gib sie als ein zum Rechteck ergaenztes Feld von
     * Textzeichen aus.
     * 
     * Codierung:
     * WAND ist ein Stueck undurchdringliche Wand.
     * LEERRAUM ist ein Leerraum.
     * EINGANG ist der Ein- und Ausgang.
     * MARKENORD ist eine Marke, die nach Norden zeigt.
     * MARKEWEST  ist eine Marke, die nach Westen zeigt.
     * MARKESUED ist eine Marke, die nach Sueden zeigt.
     * MARKEOST ist eine Marke, die nach Osten zeigt.
     * POSITIONNORD ist eine gespeicherte Position mit Blickrichtung Norden.
     * POSITIONWEST ist eine gespeicherte Position mit Blickrichtung Westen.
     * POSITIONSUED ist eine gespeicherte Position mit Blickrichtung Sueden.
     * POSITIONOST ist eine gespeicherte Position mit Blickrichtung Osten.
     * FEIND ist ein Komplize des Taeters.
     * ZEUGE1 bis ZEUGE9 ist ein Zeuge.
     * LEICHE ist ein Opfer.
     * TAETER ist der Taeter.
     * 
     * Der Rand des Rechtecks wird als Wand interpretiert.
     * Das Rechteck ist so breit wie seine laengste Zeile. Zu
     * kurze Zeilen werden mit Leerraeumen ergaenzt. Sollte
     * mehr als ein EINGANG-Zeichen enthalten sein, so werden
     * alle EINGANG-Zeichen bis auf das erste durch Leerraeume
     * ersetzt. Sollte kein EINGANG-Zeichen vorhanden sein, so
     * wird der erste Leerraum als solches verwendet. Sollte
     * kein Leerraum vorhanden sein, so wird die erste Wand
     * dafuer verwendet. Sollte auch keine Wand vorhanden sein,
     * so wird der erste Feind oder Zeuge oder die erste
     * gespeicherte Position, der erste Taeter oder das erste
     * Opfer zum Eingang. Im aeussersten Fall, dass die Datei
     * kein einziges Zeichen enthaelt, lautet die Ausgabe
     * {{EINGANG}}.
     */
    public Labyrinth lies (File labyrinthDatei)
    {
        try
        {
            /*
             * Oeffne einen Leser fuer die Datei.
             */
            BufferedReader leser = new BufferedReader (new FileReader (labyrinthDatei));

            /*
             * Speichere die Eingabezeilen zunaechst in einer Liste.
             */
            ArrayList<String> eingabezeilen = new ArrayList<String> ();

            /*
             * Merke Dir die hoechste vorkommende Breite.
             */
            int breite = 0;

            /*
             * Bestimme die Nummer der Trennzeile.
             */
            int trennzeilennummer = 0;
            boolean trennerGefunden = false;

            /*
             * Lies alle Zeilen der Datei Stueck fuer Stueck ein.
             */
            while (leser.ready ())
            {
                String zeile = leser.readLine ();

                if (zeile.equals (TRENNZEILE))
                {
                    trennerGefunden = true;
                }

                if (!trennerGefunden)
                {
                    if (zeile.length () > breite)
                    {
                        breite = zeile.length ();
                    }

                    trennzeilennummer++;
                }

                eingabezeilen.add (zeile);
            }

            leser.close ();

            /*
             * Verlaengere alle zu kurzen Kartenzeilen auf die
             * maximale Breite.
             */
            String leerkette = "";

            if (0 == breite)
            {
                breite = 1;
            }

            for (int j = 0; j < breite; j++)
            {
                leerkette = leerkette + ".";
            }

            for (int i = 0; i < trennzeilennummer; i++)
            {
                eingabezeilen.set (i, (eingabezeilen.get (i) + leerkette).substring (0, breite));
            }

            /*
             * Ermittle die Hoehe des Labyrinths.
             */
            int hoehe = trennzeilennummer;

            /*
             * Erstelle das zweidimensionale Zeichenkettenfeld,
             * das zurueckgegeben werden soll. Wenn es keine
             * Eingabezeilen gibt, erstelle ein Feld, das
             * einfach nur einen Eingang enthaelt.
             */
            char ausgabe[][] = {{EINGANG}};

            if (hoehe > 0)
            {
                ausgabe = new char[hoehe][breite];
            }

            int i = 0;
            int j;

            for (int k = 0; k < trennzeilennummer; k++)
            {
                ausgabe[i++] = eingabezeilen.get (k).toCharArray ();
            }

            /*
             * Zaehle die einzelnen Codezeichen im Labyrinth.
             */
            int eingaenge    = 0;
            int waende       = 0;
            int leerraeume   = 0;
            int markennord   = 0;
            int markenwest   = 0;
            int markensued   = 0;
            int markenost    = 0;
            int positionnord = 0;
            int positionwest = 0;
            int positionsued = 0;
            int positionost  = 0;
            int feinde       = 0;
            int zeugen1      = 0;
            int zeugen2      = 0;
            int zeugen3      = 0;
            int zeugen4      = 0;
            int zeugen5      = 0;
            int zeugen6      = 0;
            int zeugen7      = 0;
            int zeugen8      = 0;
            int zeugen9      = 0;
            int leichen      = 0;
            int taeter       = 0;

            for (i = 0; i < hoehe; i++)
            {
                for (j = 0; j < breite; j++)
                {
                    switch (ausgabe[i][j])
                    {
                        case EINGANG:
                        if (0 == eingaenge)
                        {
                            eingaenge++;
                        }
                        else
                        {
                            ausgabe[i][j] = LEERRAUM;
                            leerraeume++;
                        }
                        break;

                        case LEERRAUM:
                        leerraeume++;
                        break;

                        case WAND:
                        waende++;
                        break;

                        // Wenn eine abgespeicherte Position enthalten ist, gib sie aus.
                        case POSITIONWEST:
                        ausgabe[i][j] = POSITIONWEST;
                        positionwest++;
                        break;

                        case POSITIONSUED:
                        ausgabe[i][j] = POSITIONSUED;
                        positionsued++;
                        break;

                        case POSITIONOST:
                        ausgabe[i][j] = POSITIONOST;
                        positionost++;
                        break;

                        case POSITIONNORD:
                        ausgabe[i][j] = POSITIONNORD;
                        positionnord++;
                        break;

                        /*
                         * Die gespeicherten Karten sind nach Westen orientiert.
                         * Darum muessen wir die Marken entsprechend uebersetzen.
                         */

                        case MARKENORD:
                        ausgabe[i][j] = MARKEWEST;
                        markenwest++;
                        break;

                        case MARKEWEST:
                        ausgabe[i][j] = MARKESUED;
                        markensued++;
                        break;

                        case MARKESUED:
                        ausgabe[i][j] = MARKEOST;
                        markenost++;
                        break;

                        case MARKEOST:
                        ausgabe[i][j] = MARKENORD;
                        markennord++;
                        break;

                        case FEIND:
                        feinde++;
                        break;

                        case ZEUGE1:
                        zeugen1++;
                        break;

                        case ZEUGE2:
                        zeugen2++;
                        break;

                        case ZEUGE3:
                        zeugen3++;
                        break;

                        case ZEUGE4:
                        zeugen4++;
                        break;

                        case ZEUGE5:
                        zeugen5++;
                        break;

                        case ZEUGE6:
                        zeugen6++;
                        break;

                        case ZEUGE7:
                        zeugen7++;
                        break;

                        case ZEUGE8:
                        zeugen8++;
                        break;

                        case ZEUGE9:
                        zeugen9++;
                        break;

                        case LEICHE:
                        leichen++;
                        break;

                        case TAETER:
                        taeter++;
                        break;

                        default:
                        ausgabe[i][j] = LEERRAUM;
                        leerraeume++;
                        break;
                    }
                }
            }

            /*
             * Sorge dafuer, dass es einen Eingang gibt.
             */
            if (0 == eingaenge)
            {
                char suchzeichen;

                if (0 != leerraeume)
                {
                    suchzeichen = LEERRAUM;
                }
                else if (0 != markennord)
                {
                    suchzeichen = MARKENORD;
                }
                else if (0 != markenwest)
                {
                    suchzeichen = MARKEWEST;
                }
                else if (0 != markensued)
                {
                    suchzeichen = MARKESUED;
                }
                else if (0 != markenost)
                {
                    suchzeichen = MARKEOST;
                }
                else if (0 != waende)
                {
                    suchzeichen = WAND;
                }
                else if (0 != feinde)
                {
                    suchzeichen = FEIND;
                }
                else if (0 != zeugen1)
                {
                    suchzeichen = ZEUGE1;
                }
                else if (0 != zeugen2)
                {
                    suchzeichen = ZEUGE2;
                }
                else if (0 != zeugen3)
                {
                    suchzeichen = ZEUGE3;
                }
                else if (0 != zeugen4)
                {
                    suchzeichen = ZEUGE4;
                }
                else if (0 != zeugen5)
                {
                    suchzeichen = ZEUGE5;
                }
                else if (0 != zeugen6)
                {
                    suchzeichen = ZEUGE6;
                }
                else if (0 != zeugen7)
                {
                    suchzeichen = ZEUGE7;
                }
                else if (0 != zeugen8)
                {
                    suchzeichen = ZEUGE8;
                }
                else if (0 != zeugen9)
                {
                    suchzeichen = ZEUGE9;
                }
                else if (0 != positionnord)
                {
                    suchzeichen = POSITIONNORD;
                }
                else if (0 != positionwest)
                {
                    suchzeichen = POSITIONWEST;
                }
                else if (0 != positionsued)
                {
                    suchzeichen = POSITIONSUED;
                }
                else if (0 != positionost)
                {
                    suchzeichen = POSITIONOST;
                }
                else if (0 != taeter)
                {
                    suchzeichen = TAETER;
                }
                else
                {
                    suchzeichen = LEICHE;
                }

                boolean erledigt = false;

                for (i = 0; i < hoehe; i++)
                {
                    for (j = 0; j < breite; j++)
                    {
                        if (!erledigt && ausgabe[i][j] == suchzeichen)
                        {
                            ausgabe[i][j] = EINGANG;
                            erledigt = true;
                        }
                    }
                }
            }

            Labyrinth ausgabeLabyrinth;
            ausgabeLabyrinth = new Labyrinth ();
            ausgabeLabyrinth.setzeKarte (ausgabe);

            /*
             * Lies die Zeugenbemerkungen aus den Eingabezeilen ab der
             * Trennzeile aus.
             */
            int    aktuellerZeuge = 0;
            String aktuelleZeile;
            String gesammelteZeilen = "";
            char   erstesZeichen;
            char   zweitesZeichen;

            for (int k = trennzeilennummer; k < eingabezeilen.size (); k++)
            {
                /*
                 * Bestimme das erste und zweite Textzeichen der aktuellen
                 * Zeile.
                 */
                aktuelleZeile  = eingabezeilen.get (k);

                if (aktuelleZeile.length () > 0)
                {
                    erstesZeichen = aktuelleZeile.charAt (0);
                }
                else
                {
                    erstesZeichen = 0;
                }

                if (aktuelleZeile.length () > 1)
                {
                    zweitesZeichen = aktuelleZeile.charAt (1);
                }
                else
                {
                    zweitesZeichen = 0;
                }

                /*
                 * Wenn das erste Textzeichen zwischen '1' und '9' liegt und das zweite
                 * Textzeichen ein Doppelpunkt ist, speichere die letzte Zeugenaussage
                 * und beginne eine neue Zeugenaussage, wobei das erste Zeichen angibt,
                 * wie der aktuelle Zeuge lautet.
                 */

                if (erstesZeichen >= '1' && erstesZeichen <= '9' && zweitesZeichen == ':')
                {
                    if (aktuellerZeuge < 10)
                    {
                        ausgabeLabyrinth.setzeZeugenaussage (aktuellerZeuge, gesammelteZeilen);
                    }
                    else if (aktuellerZeuge == 10)
                    {
                        ausgabeLabyrinth.setzeEinleitungstext (gesammelteZeilen);
                    }

                    aktuellerZeuge   = (int) (erstesZeichen - '0');
                    
                    if (aktuelleZeile.length () > 2)
                    {
                        gesammelteZeilen = aktuelleZeile.substring (2);
                    }
                    else
                    {
                        gesammelteZeilen = "";
                    }
                }
                else if (erstesZeichen == 'e' && zweitesZeichen == ':')
                {
                    if (aktuellerZeuge < 10)
                    {
                        ausgabeLabyrinth.setzeZeugenaussage (aktuellerZeuge, gesammelteZeilen);
                    }

                    else if (aktuellerZeuge == 10)
                    {
                        ausgabeLabyrinth.setzeEinleitungstext (gesammelteZeilen);
                    }

                    aktuellerZeuge = 10;
                    
                    if (aktuelleZeile.length () > 2)
                    {
                        gesammelteZeilen = aktuelleZeile.substring (2);
                    }
                    else
                    {
                        gesammelteZeilen = "";
                    }
                }
                else
                {
                    gesammelteZeilen = gesammelteZeilen + aktuelleZeile;
                }

            }

            if (aktuellerZeuge < 10)
            {
                ausgabeLabyrinth.setzeZeugenaussage (aktuellerZeuge, gesammelteZeilen);
            }
            else if (aktuellerZeuge == 10)
            {
                ausgabeLabyrinth.setzeEinleitungstext (gesammelteZeilen);
            }

            return ausgabeLabyrinth;
        }
        catch (Exception fehler)
        {
            return null;
        }
    }
}
